/* Contains the worm's internal data structure (a circular buffer)
   and code for deciding on the position and compass direction
   of the next worm move.
 */

import java.awt.*;
import java.awt.geom.*;

public class Worm {
	// Madhsia dhe numri i pikave n krimbin
	private static final int DOTSIZE = 12;
	private static final int RADIUS = DOTSIZE / 2;
	private static final int MAXPOINTS = 40;

	// Konstantat e drejtimit t lvizjes/busolls
	private static final int NUM_DIRS = 8;
	private static final int N = 0; // veri... n drejtim t akrepave
	private static final int NE = 1;
	private static final int E = 2;
	private static final int SE = 3;
	private static final int S = 4;
	private static final int SW = 5;
	private static final int W = 6;
	private static final int NW = 7;

	private int currCompass; // ruan drejtimin aktual

	// Ruan inkrementet pr secilin nga drejtimet
	// Nj inkrement i shtohet pozits s vjetr t koks
	// pr t fituar pozitn e re.
	Point2D.Double incrs[];

	// Informata mbi probabilitetin pr t zgjedhur nj drejtim lvizjeje
	private static final int NUM_PROBS = 9;
	private int probsForOffset[];

	// cells[] ruan pikat q prbjn krimbin
	// trajtohet si bafer cirkular
	private Point cells[];
	private int nPoints;
	private int tailPosn, headPosn; // bishti dhe koka e baferit

	private int pWidth, pHeight; // dimensionet e panelit
	private long startTime; // n ms
	private Obstacles obs;

	public Worm(int pW, int pH, Obstacles os) {
		pWidth = pW;
		pHeight = pH;
		obs = os;
		cells = new Point[MAXPOINTS]; // inicializo baferin
		nPoints = 0;
		headPosn = -1;
		tailPosn = -1;

		// inkrementet pr secilin drejtim
		incrs = new Point2D.Double[NUM_DIRS];
		incrs[N] = new Point2D.Double(0.0, -1.0);
		incrs[NE] = new Point2D.Double(0.7, -0.7);
		incrs[E] = new Point2D.Double(1.0, 0.0);
		incrs[SE] = new Point2D.Double(0.7, 0.7);
		incrs[S] = new Point2D.Double(0.0, 1.0);
		incrs[SW] = new Point2D.Double(-0.7, 0.7);
		incrs[W] = new Point2D.Double(-1.0, 0.0);
		incrs[NW] = new Point2D.Double(-0.7, -0.7);

		// informata mbi probabilitetin e zgjedhjes s nj drejtimi
		// 0 = pa ndryshim, -1 kthim kundr akrepave,
		// 1 kthim nga akrepat, -2 kthim i ashpr kundr akrepave etj.
		/*
		 * Shprndarja e probabilitetit bn q krimbi zakonisht vazhdon drejt
		 * por mund t kthehet pak majtas ose pak djathtas.
		 */
		probsForOffset = new int[NUM_PROBS];
		probsForOffset[0] = 0;
		probsForOffset[1] = 0;
		probsForOffset[2] = 0;
		probsForOffset[3] = 1;
		probsForOffset[4] = 1;
		probsForOffset[5] = 2;
		probsForOffset[6] = -1;
		probsForOffset[7] = -1;
		probsForOffset[8] = -2;

	}

	/* A sht (x,y) afr koks? */
	public boolean nearHead(int x, int y) {
		if (nPoints > 0) {
			if ((Math.abs(cells[headPosn].x + RADIUS - x) <= DOTSIZE)
					&& (Math.abs(cells[headPosn].y + RADIUS - y) <= DOTSIZE))
				return true;
		}
		return false;
	}

	/* A sht (x,y) afr afr ndonj pjese t trupit? */
	public boolean touchedAt(int x, int y) {
		int i = tailPosn;
		while (i != headPosn) {
			if ((Math.abs(cells[i].x + RADIUS - x) <= RADIUS)
					&& (Math.abs(cells[i].y + RADIUS - y) <= RADIUS))
				return true;
			i = (i + 1) % MAXPOINTS;
		}
		return false;
	}

	/*
	 * A move causes the addition of a new dot to the front of the worm, which
	 * becomes its new head. A dot has a position and compass direction/bearing,
	 * which is derived from the position and bearing of the old head.
	 * 
	 * move() is complicated by having to deal with 3 cases: when the worm is
	 * first created when the worm is growing when the worm is MAXPOINTS long
	 * (then the addition of a new head must be balanced by the removal of a
	 * tail dot)
	 */
	public void move() {
		int prevPosn = headPosn; // ruaj pozitn e vjetr t koks
		headPosn = (headPosn + 1) % MAXPOINTS;

		if (nPoints == 0) { // vargu bosh n fillim
			tailPosn = headPosn;
			currCompass = (int) (Math.random() * NUM_DIRS); // drejt. i rast.
			cells[headPosn] = new Point(pWidth / 2, pHeight / 2); // qendra
			nPoints++;
		} else if (nPoints == MAXPOINTS) { // vargu sht plot
			tailPosn = (tailPosn + 1) % MAXPOINTS; // largo bishtin
			newHead(prevPosn);
		} else { // vargu nuk sht plot
			newHead(prevPosn);
			nPoints++;
		}
	}

	/*
	 * Create new head position and compass direction/bearing.
	 * 
	 * This has two main parts. Initially we try to generate a head by varying
	 * the old position/bearing. But if the new head hits an obstacle, then we
	 * shift to a second phase.
	 * 
	 * In the second phase we try a head which is 90 degrees clockwise, 90
	 * degress clockwise, or 180 degrees reversed so that the obstacle is
	 * avoided. These bearings are stored in fixedOffs[].
	 */
	private void newHead(int prevPosn) {
		Point newPt;
		int newBearing;
		int fixedOffs[] = { -2, 2, -4 }; // offset-et pr t evituar penges

		newBearing = varyBearing();
		newPt = nextPoint(prevPosn, newBearing);
		// Nj pozit e re mbshtetur n pozitn paraprake
		// dhe ndryshim semi-t rastsishm drejtimi

		if (obs.hits(newPt, DOTSIZE)) {
			for (int i = 0; i < fixedOffs.length; i++) {
				newBearing = calcBearing(fixedOffs[i]);
				newPt = nextPoint(prevPosn, newBearing);
				if (!obs.hits(newPt, DOTSIZE))
					break; // nj nga offset-et e fiksuara bn pun
			}
		}
		cells[headPosn] = newPt; // pozita e re e koks
		currCompass = newBearing; // drejtimi i ri i lvizjes
	}

	/* Ndrysho drjetimin e lvizjes n mnyr semi-t rastsishme */
	private int varyBearing() {
		int newOffset = probsForOffset[(int) (Math.random() * NUM_PROBS)];
		return calcBearing(newOffset);
	}

	/* Prdor offset-in pr t llogaritur drejtimin e ri */
	private int calcBearing(int offset) {
		int turn = currCompass + offset;
		// ensure that turn is between N to NW (0 to 7)
		if (turn >= NUM_DIRS)
			turn = turn - NUM_DIRS;
		else if (turn < 0)
			turn = NUM_DIRS + turn;
		return turn;
	}

	/*
	 * Return the next coordinate based on the previous position and a compass
	 * bearing.
	 * 
	 * Convert the compass bearing into predetermined increments (stored in
	 * incrs[]). Add the increments multiplied by the DOTSIZE to the old head
	 * position. Deal with wraparound.
	 */
	private Point nextPoint(int prevPosn, int bearing) {
		// merr inkrementet pr drejtimin e lvizjes
		Point2D.Double incr = incrs[bearing];

		int newX = cells[prevPosn].x + (int) (DOTSIZE * incr.x);
		int newY = cells[prevPosn].y + (int) (DOTSIZE * incr.y);

		// modifiko newX/newY n qoft se jasht pluhrs
		if (newX + DOTSIZE < 0) // jasht skajit t majt?
			newX = newX + pWidth;
		else if (newX > pWidth) // jasht skajit t djatht?
			newX = newX - pWidth;

		if (newY + DOTSIZE < 0) // jasht skajit t siprm?
			newY = newY + pHeight;
		else if (newY > pHeight) // jasht skajit t poshtm?
			newY = newY - pHeight;

		return new Point(newX, newY);
	}

	/* Vizato nj krimb t zi me kok t kuqe */
	public void draw(Graphics g) {
		if (nPoints > 0) {
			g.setColor(Color.black);
			int i = tailPosn;
			while (i != headPosn) {
				g.fillOval(cells[i].x, cells[i].y, DOTSIZE, DOTSIZE);
				i = (i + 1) % MAXPOINTS;
			}
			g.setColor(Color.red);
			g.fillOval(cells[headPosn].x, cells[headPosn].y, DOTSIZE, DOTSIZE);
		}
	}
}
