/* The game's drawing surface. It shows:
 - the moving worm
 - the obstacles (blue boxes)
 - the current average FPS
 */

import javax.swing.*;
import java.awt.event.*;
import java.awt.*;
import java.text.DecimalFormat;

public class WormPanel extends JPanel implements Runnable {
	private static final int PWIDTH = 500; // madhsia e panelit
	private static final int PHEIGHT = 400;

	private long gameStartTime;
	private int timeSpentInGame = 0; // n skonda

	private long frameCount = 0;
	private double averageFPS = 0.0;

	private DecimalFormat df = new DecimalFormat("0.##"); // 2 decimale

	private Thread animator; // fija q performon animimin
	private boolean running = false; // pr t ndaluar fijen e animimit
	private boolean isPaused = false;

	private long period; // periudha ndrjet vizatimeve n ms

	private WormChase wcTop;
	private Worm fred; // krimbi
	private Obstacles obs; // pengesat

	// prdoren kur loja t terminohet
	private boolean gameOver = false;
	private int score = 0;
	private Font font;
	private FontMetrics metrics;

	// renderimi off-screen
	private Graphics dbg;
	private Image dbImage = null;

	public WormPanel(WormChase wc, long period) {
		wcTop = wc;
		this.period = period;

		setBackground(Color.white);
		setPreferredSize(new Dimension(PWIDTH, PHEIGHT));

		setFocusable(true);
		requestFocus(); // JPanel ka fokusin, prano ngjarje tastesh
		readyForTermination();

		// krijo komponentat e lojs
		obs = new Obstacles(wcTop);
		fred = new Worm(PWIDTH, PHEIGHT, obs);

		addMouseListener(new MouseAdapter() {
			public void mousePressed(MouseEvent e) {
				testPress(e.getX(), e.getY());
			}
		});

		// vj fontin e mesazhit
		font = new Font("SansSerif", Font.BOLD, 24);
		metrics = this.getFontMetrics(font);
	}

	private void readyForTermination() {
		addKeyListener(new KeyAdapter() {
			// dgjo pr esc, q, end, ctrl-c mbi plhurn
			// pr t lejuar dalje
			public void keyPressed(KeyEvent e) {
				int keyCode = e.getKeyCode();
				if ((keyCode == KeyEvent.VK_ESCAPE)
						|| (keyCode == KeyEvent.VK_Q)
						|| (keyCode == KeyEvent.VK_END)
						|| ((keyCode == KeyEvent.VK_C) && e.isControlDown())) {
					running = false;
				}
			}
		});
	}

	/*
	 * Prit q ky JPanel t shtohet te JFrame/JApplet para startimit.
	 */
	public void addNotify() {
		super.addNotify(); // krijon peer-in
		startGame(); // starton fijen
	}

	/* Inicializo dhe starto fijen */
	private void startGame() {
		if (animator == null || !running) {
			animator = new Thread(this);
			animator.start();
		}
	}

	// ------------- Metodat e ciklit t jets s lojs ------------
	// t invokuara nga metodat window listener t JFrame

	/* Invokohet kur JFrame t aktivizoet / deikonifikohet */
	public synchronized void resumeGame() {
		isPaused = false;
		notify();
	}

	/* Invokohet kur JFrame t deaktivizohet / ikonifikohet */
	public void pauseGame() {
		isPaused = true;
	}

	/* Invokohet nga JFrame sht duke u mbyllur */
	public synchronized void stopGame() {
		running = false;
		notify();
	}

	// ----------------------------------------------

	/* A sht (x,y) afr koks apo t shtohet nj penges? */
	private void testPress(int x, int y) {
		if (!isPaused && !gameOver) {
			if (fred.nearHead(x, y)) {
				gameOver = true;
				score = (40 - timeSpentInGame) + (40 - obs.getNumObstacles());
			} else { // shto nj penges nse ka mundsi
				if (!fred.touchedAt(x, y))
					obs.add(x, y);
			}
		}
	}

	/* Cikli i animimit */
	public void run() {
		gameStartTime = System.currentTimeMillis();

		long beforeTime, timeDiff, sleepTime;
		beforeTime = System.currentTimeMillis();
		running = true;
		while (running) {
			try {
				if (isPaused) {
					synchronized (this) {
						while (isPaused && running)
							wait();
					}
				}
			} catch (InterruptedException e) {
			}
			gameUpdate();
			gameRender();
			paintScreen();
			timeDiff = System.currentTimeMillis() - beforeTime;
			sleepTime = period - timeDiff; // koha e mbetur n iterimin
			if (sleepTime <= 0) // update-render zgjati m tepr
				sleepTime = 5; // megjithat flej pak
			try {
				Thread.sleep(sleepTime); // n ms
			} catch (InterruptedException ex) {
			}
			beforeTime = System.currentTimeMillis();
			storeStats();
		}
		printStats();
		System.exit(0); // kshtuq dritarja zhduket
	}

	private void gameUpdate() {
		if (!isPaused && !gameOver)
			fred.move();
	}

	private void gameRender() {
		if (dbImage == null) {
			dbImage = createImage(PWIDTH, PHEIGHT);
			if (dbImage == null) {
				System.out.println("dbImage is null");
				return;
			} else
				dbg = dbImage.getGraphics();
		}

		dbg.setColor(Color.white);
		dbg.fillRect(0, 0, PWIDTH, PHEIGHT);

		dbg.setColor(Color.blue);
		dbg.setFont(font);

		// raporto numrin e kornizave & FPS mesatar
		// dbg.drawString("Frame Count " + frameCount, 10, 25);
		dbg.drawString("Average FPS: " + df.format(averageFPS), 20, 25);

		dbg.setColor(Color.black);

		// vizato elementet e lojs
		obs.draw(dbg);
		fred.draw(dbg);

		if (gameOver)
			gameOverMessage(dbg);
	}

	/* qendrso mesazhin game-over n panelin */
	private void gameOverMessage(Graphics g) {
		String msg = "Game Over. Rezultati juaj: " + score;
		int x = (PWIDTH - metrics.stringWidth(msg)) / 2;
		int y = (PHEIGHT - metrics.getHeight()) / 2;
		g.setColor(Color.red);
		g.setFont(font);
		g.drawString(msg, x, y);
	}

	/* Vj n ekran imazhin e baferuar */
	private void paintScreen() {
		Graphics g;
		try {
			g = this.getGraphics();
			if ((g != null) && (dbImage != null))
				g.drawImage(dbImage, 0, 0, null);
			g.dispose();
		} catch (Exception e) {
			System.out.println("Graphics context error: " + e);
		}
	}

	/*
	 * Statistikat
	 */
	private void storeStats() {
		frameCount++;

		long timeNow = System.currentTimeMillis();
		timeSpentInGame = (int) ((timeNow - gameStartTime) / 1000); // ms -->
																	// secs
		wcTop.setTimeSpent(timeSpentInGame);

		averageFPS = frameCount / ((timeNow - gameStartTime) / 1000.0);
	}

	private void printStats() {
		System.out.println("Korniza: " + frameCount);
		System.out.println("Mesatarisht FPS: " + df.format(averageFPS));
		System.out.println("Koha: " + timeSpentInGame + " secs");
		System.out.println("Kuti: " + obs.getNumObstacles());
	}

}
